# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2020 Ledger SAS, written by Olivier Hériveaux


from typing import Optional
import serial.serialutil
from binascii import hexlify
from .exceptions import ConnectionFailure, ProtocolError, VersionNotSupported
from .stage import Stage
from .vector import Vector
from enum import Enum


class M3FSCommand(Enum):
    """
    Command integer ID for New Scale Technologies M3-FS.
    https://www.newscaletech.com/downloads/please-register/
    """

    READ_FIRM_VERSION = 1
    MOVE_TO_TARGET = 8
    VIEW_CLOSED_LOOP_STATUS_POS = 10


class M3FS(Stage):
    """
    Class to command New Scale Technologies M3-FS focus modules in serial mode
    (VCP).
    If a M3-FS device is seen as a USBXpress chip, it may be used as a
    Virtual-Com-Port after a few cumbersome operations (custom driver with
    correct VID/DID).
    """

    def __init__(self, dev: Optional[str] = None, baudrate=250000):
        """
        Connect to the device. If the serial device cannot be opened, a
        ConnectionFailure exception is thrown. If the device version is not
        supported, a VersionNotSupported error is thrown.

        :param dev: Serial device. For instance `'/dev/ttyUSB0'`.
            If not provided, a suitable device is searched according to
            according to vendor and product IDs
        :param baudrate: Serial baudrate.
        """
        super().__init__()
        self.resolution_um = 0.5
        try:
            self.serial = serial.Serial(
                dev, baudrate=baudrate, stopbits=serial.STOPBITS_TWO
            )
        except serial.serialutil.SerialException as e:
            raise ConnectionFailure() from e
        # Test version string.
        # Currently, only one version has been tested. Some others may be added
        # in the future...
        self.serial.timeout = 1
        try:
            res = self.command(M3FSCommand.READ_FIRM_VERSION)
        except ProtocolError as e:
            raise ConnectionFailure() from e
        self.serial.timeout = None
        if res != "1 VER 4.7.3 M3-FS":
            raise VersionNotSupported(res)

    def __send(self, command: M3FSCommand, data: Optional[str] = None):
        """
        Send a command to the controller.

        :param command: Command integer ID.
        :param data: Extra data string.
        """
        if data is not None:
            assert ("<" not in data) and (">" not in data) and ("\r" not in data)
        if command.value not in range(100):
            raise ValueError("Invalid command ID.")
        full_command = "<{0:02d}".format(command.value)
        if data is not None:
            full_command += " " + data
        full_command += ">\r"
        self.serial.write(full_command.encode())

    def __receive(self):
        """
        Read next response from the controller.

        :return: Response string, without '<', '>' and carriage return.
        """
        # Get start '<'
        b = self.serial.read(1)
        if len(b) != 1:
            raise ProtocolError()
        if b[0] != ord("<"):
            raise ProtocolError()
        # Get the response bytes until '>'
        result = bytearray()
        while True:
            b = self.serial.read(1)
            if len(b) != 1:
                raise ProtocolError()
            b = b[0]
            if b == ord(">"):
                break
            else:
                if b == ord("\r"):
                    raise ProtocolError()
                result.append(b)
        # Get carriage return
        b = self.serial.read(1)
        if len(b) != 1:
            raise ProtocolError()
        if b[0] != ord("\r"):
            raise ProtocolError()
        return result.decode()

    def command(self, command: M3FSCommand, data=None):
        """
        Send a command to the controller and get the response.

        :param command: Command integer ID.
        :param data: Extra data string.
        :return: Response data string.
        """
        self.__send(command, data)
        res = self.__receive()
        # Check that the command id in the response is the same as the command
        # id in the request.
        if (len(res) < 2) or (int(res[0:2]) != command.value):
            raise ProtocolError(
                f"Unexpected response after sending command {command}:", res
            )
        if len(res) == 2:
            return None
        else:
            if res[2] != " ":
                raise ProtocolError(
                    f"Unexpected response after sending command {command}:", res
                )
            return res[3:]

    def __get_closed_loop_status(self):
        """
        Query closed-loop status and position.

        :return: Tuple of 3 int.
        """
        command = M3FSCommand.VIEW_CLOSED_LOOP_STATUS_POS
        res = self.command(command)
        if res is None:
            raise ProtocolError(
                f"Unexpected response after sending command {command}: Got response without data."
            )
        res = list(bytes.fromhex(x) for x in res.split(" "))
        if len(res) != 3:
            raise ProtocolError(
                f"Unexpected response after sending command {command}: Expecting 3 values, got {res}."
            )
        motor_status = int.from_bytes(res[0], "big", signed=False)
        position = int.from_bytes(res[1], "big", signed=True)
        error = int.from_bytes(res[2], "big", signed=True)
        return motor_status, position, error

    @property
    def position(self) -> Vector:
        """
        Stage position, in micrometers.

        :getter: Query and return current stage position.
        :setter: Move stage. Wait until position is reached.
        """
        _, position, _ = self.__get_closed_loop_status()
        return Vector(position * self.resolution_um)

    @position.setter
    def position(self, value: Vector):
        # To check dimension and range of the given value
        pos_setter = Stage.position.fset
        assert pos_setter is not None
        pos_setter(self, value)

        val = round(value.x / self.resolution_um).to_bytes(4, "big", signed=True)
        self.command(M3FSCommand.MOVE_TO_TARGET, hexlify(val).decode())
        # Now wait until motor is not moving anymore
        while self.is_moving:
            pass

    @property
    def is_moving(self) -> bool:
        return bool(self.__get_closed_loop_status()[0] & 4)



#!/bin/python3
from typing import Optional
from PyQt6.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QGridLayout,
    QLabel,
    QComboBox,
    QPushButton,
    QCheckBox,
    QLineEdit,
)
from PyQt6.QtCore import QObject, QTimer, QLocale, QCoreApplication
from PyQt6.QtGui import QDoubleValidator
from ..cncrouter import CNCRouter
from ..corvus import Corvus
from ..smc100 import SMC100
from ..stage import Stage
from ..m3fs import M3FS
from enum import Enum
from serial.tools.list_ports import comports
from serial.tools.list_ports_common import ListPortInfo


class StageType(str, Enum):
    CNC = "CNC"
    Corvus = "Corvus"
    SMC = "SMC100"
    M3FS = "M3FS"


class StageWindow(QWidget):
    def set_controls_enabled(self, enabled: bool):
        for c in self.controls:
            c.setEnabled(enabled)

    def connect(self, on_off):
        if on_off:
            selected = self.stage_selection.currentText()
            port = self.port_selection.currentData()
            dev = port.device if isinstance(port, ListPortInfo) else None

            # Instanciate stage according to current stage selection
            if selected == StageType.CNC:
                self.stage = CNCRouter(dev)
            elif selected == StageType.Corvus:
                self.stage = Corvus(dev)
            elif selected == StageType.SMC:
                self.stage = SMC100(dev, [1, 2])
            elif selected == StageType.M3FS:
                self.stage = M3FS(dev, baudrate=115200)
            self.position_timer.start(100)

        else:
            del self.stage
            self.stage = None

            self.position_timer.stop()

        self.stage_selection.setDisabled(on_off)
        self.port_selection.setDisabled(on_off)

        self.set_controls_enabled(on_off)

        if self.stage is not None:
            self.stage.wait_routine = lambda: QCoreApplication.processEvents()

    def __init__(self):
        super().__init__()

        # Current stage
        self.stage: Optional[Stage] = None

        # This flag is used to limit the communication
        # with the stage by not making updates of the position
        self.in_motion = False

        # Moving controls
        self.controls = []

        vbox = QVBoxLayout()
        self.setLayout(vbox)

        box = QHBoxLayout()
        vbox.addLayout(box)
        w = QLabel("Stage Selection")
        box.addWidget(w)
        self.stage_selection = w = QComboBox()
        w.addItems([StageType.CNC, StageType.Corvus, StageType.SMC, StageType.M3FS])
        box.addWidget(w)
        self.port_selection = w = QComboBox()
        w.addItem("Auto detection", None)
        for port in comports():
            d = port.device
            if port.vid and port.pid:
                d += f" -- {port.vid:04x}:{port.pid:04x}"
            if port.product:
                d += f" -- {port.product}"
            if port.serial_number:
                d += f" -- {port.serial_number}"
            w.addItem(d, userData=port)
        box.addWidget(w)
        self.connect_button = w = QPushButton("Connect")
        w.setCheckable(True)
        w.clicked.connect(self.connect)
        box.addWidget(w)

        box = QHBoxLayout()
        vbox.addLayout(box)
        box.addWidget(QLabel("Step"))
        self.step_selection = w = QComboBox()
        for step in [50, 10, 5, 1, 0.5, 0.1]:
            w.addItem(f"{step} mm", step)
        w.setCurrentIndex(3)
        box.addWidget(w)
        self.controls.append(w)

        self.moving_grid = grid = QGridLayout()
        w = QPushButton("Y+")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 0, 1)
        w = QPushButton("X-")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 1, 0)
        w = QPushButton("Y-")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 2, 1)
        w = QPushButton("X+")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 1, 2)
        w = QPushButton("Z+")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 0, 3)
        w = QPushButton("Z-")
        w.clicked.connect(self.bouton_moved)
        self.controls.append(w)
        grid.addWidget(w, 2, 3)
        w = QPushButton("Home")
        w.clicked.connect(self.home)
        self.controls.append(w)
        grid.addWidget(w, 1, 1)
        vbox.addLayout(grid)

        box = QHBoxLayout()
        vbox.addLayout(box)
        self.z_offset = w = QCheckBox("Z offset (mm)")
        w.setChecked(True)
        box.addWidget(w)
        self.controls.append(w)
        self.z_offset_sel = w = QComboBox()
        for value in [10, 5, 1, 0.5, 0.1]:
            w.addItem(f"{value} mm", value)
        w.setCurrentIndex(2)
        box.addWidget(w)
        self.controls.append(w)

        v = QDoubleValidator()
        v.setLocale(QLocale.c())
        box = QHBoxLayout()
        vbox.addLayout(box)
        self.go_x = w = QLineEdit()
        w.setValidator(v)
        box.addWidget(w)
        self.go_y = w = QLineEdit()
        w.setValidator(v)
        box.addWidget(w)
        self.go_z = w = QLineEdit()
        w.setValidator(v)
        box.addWidget(w)
        w = QPushButton("Go to position")
        w.clicked.connect(self.go_to_position)
        self.controls.append(w)
        box.addWidget(w)

        # Disable all controls
        self.set_controls_enabled(False)

        self.position_label = w = QLabel("Pos")
        vbox.addWidget(w)
        self.position_timer = QTimer()
        self.position_timer.timeout.connect(self.update_position)

    def update_position(self):
        if self.stage is None or self.in_motion:
            return
        position = self.stage.position
        self.position_label.setText(",".join([f"{i:.02f}" for i in position.data]))

    def bouton_moved(self):
        if self.stage is None:
            return
        self.in_motion = True
        button = QObject().sender()
        assert isinstance(button, QPushButton)
        axe, direction = button.text()
        axe = {"X": 0, "Y": 1, "Z": 2}[axe]
        direction = {"+": 1.0, "-": -1.0}[direction]
        step = direction * self.step_selection.currentData()
        pos = self.stage.position
        z_offset = (
            self.z_offset_sel.currentData()
            if axe != 2 and self.z_offset.isChecked()
            else None
        )

        if z_offset is not None:
            pos[2] += z_offset
            self.stage.move_to(pos)

        pos[axe] += step
        self.stage.move_to(pos)

        if z_offset is not None:
            pos[2] -= z_offset
            self.stage.move_to(pos)
        self.in_motion = False

    def home(self):
        if self.stage is None:
            return
        self.stage.home(wait=True)

    def go_to_position(self):
        if self.stage is None:
            return
        x, y, z = (
            QLocale.c().toDouble(self.go_x.text())[0],
            QLocale.c().toDouble(self.go_y.text())[0],
            QLocale.c().toDouble(self.go_z.text())[0],
        )
        pos = self.stage.position
        pos.x = x
        pos.y = y
        pos.z = z
        self.stage.move_to(pos)



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018 Ledger SAS, written by Olivier Hériveaux and Manuel San Pedro


import unittest
from typing import Union


class Vector:
    """
    Some basic vector manipulation class for stages control.
    """

    def __init__(self, *args, dim=None):
        """
        :param args: Initial values of the vector.
        :param dim: Dimension of the vector.
        """
        if dim is not None:
            if len(args) > dim:
                raise ValueError("Too many initial values for this dimension.")
            self.data = list(args)
            while len(self.data) < dim:
                self.data.append(0)
        else:
            self.data = list(args)

    @property
    def x(self):
        """First element of the vector."""
        return self.data[0]

    @x.setter
    def x(self, value):
        self.data[0] = value

    @property
    def y(self):
        """Second element of the vector."""
        return self.data[1]

    @y.setter
    def y(self, value):
        self.data[1] = value

    @property
    def z(self):
        """Third element of the vector."""
        return self.data[2]

    @z.setter
    def z(self, value):
        self.data[2] = value

    @property
    def w(self):
        """Fourth element of the vector."""
        return self.data[3]

    @w.setter
    def w(self, value):
        self.data[3] = value

    @property
    def xy(self):
        """First and second elements of the vector, as a 2D Vector."""
        return Vector(self.data[0], self.data[1])

    @xy.setter
    def xy(self, value):
        self.data[0] = value.data[0]
        self.data[1] = value.data[1]

    def __getitem__(self, key):
        """
        :return: Vector element if key is an integer, or a list of items if key
            is a slice.
        :param key: int or slice.
        """
        return self.data[key]

    def __setitem__(self, key, value):
        """
        Set an item of the vector.
        :param key: int or slice.
        :param value: New value.
        """
        self.data[key] = value

    def __str__(self):
        """
        :return: String representation of the vector. For instance '(1, 2, 3)'.
        """
        return "(" + ",".join(str(x) for x in self.data) + ")"

    def __add__(self, other):
        """
        :return: Sum of this vector to the other.
        :param other: A Vector instance of same dimension.
        """
        dim = len(self)
        if len(other) != dim:
            raise ValueError("Incorrect vector size")
        result = Vector(*self.data)
        for i in range(dim):
            result.data[i] += other[i]
        return result

    def __sub__(self, other):
        """
        :return: Difference between this vector and the other.
        :param other: A Vector instance of same dimension.
        """
        dim = len(self)
        if len(other) != dim:
            raise ValueError("Incorrect vector size")
        result = Vector(*self.data)
        for i in range(dim):
            result.data[i] -= other[i]
        return result

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        """ """
        return "Vector(" + ",".join(str(x) for x in self.data) + ")"

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self[i] != other[i]:
                return False
        return True

    def __mul__(self, other: Union["Vector", int, float]):
        """
        :return: Scalar multiplication between this vector and the other.
        :param other: A Vector instance of same dimension, or an integer or a float.
        """
        dim = len(self)
        if isinstance(other, (int, float)):
            result = Vector(dim=dim)
            for i in range(dim):
                result[i] = self[i] * other
            return result

        result = Vector(dim=dim)
        if len(other) != dim:
            raise ValueError("Incorrect vector size")
        for i in range(dim):
            result[i] = self[i] * other[i]
        return result

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return self * (1.0 / other)
        else:
            raise TypeError(
                "Incorrect type for second operand. int or float is expected."
            )


class TestVector(unittest.TestCase):
    def test_init(self):
        v = Vector(1)
        self.assertEqual(len(v), 1)
        self.assertEqual(v.data, [1])
        v = Vector(1, 2)
        self.assertEqual(len(v), 2)
        self.assertEqual(v.data, [1, 2])
        v = Vector(1, 2, 3)
        self.assertEqual(len(v), 3)
        self.assertEqual(v.data, [1, 2, 3])
        v = Vector(1, 2, 3, 4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [1, 2, 3, 4])

        for i in range(5):
            v = Vector(dim=i)
            self.assertEqual(len(v), i)

        v = Vector(1, dim=1)
        self.assertEqual(len(v), 1)
        self.assertEqual(v.data, [1])
        v = Vector(1, 2, dim=2)
        self.assertEqual(len(v), 2)
        self.assertEqual(v.data, [1, 2])
        v = Vector(1, 2, 3, dim=3)
        self.assertEqual(len(v), 3)
        self.assertEqual(v.data, [1, 2, 3])
        v = Vector(1, 2, 3, 4, dim=4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [1, 2, 3, 4])

        # Test padding with 0
        v = Vector(dim=4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [0, 0, 0, 0])
        v = Vector(1, dim=4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [1, 0, 0, 0])
        v = Vector(1, 2, dim=4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [1, 2, 0, 0])
        v = Vector(1, 2, 3, dim=4)
        self.assertEqual(len(v), 4)
        self.assertEqual(v.data, [1, 2, 3, 0])

        # Test number of args checking
        with self.assertRaises(ValueError):
            _ = Vector(1, 2, 3, 4, 5, dim=4)

    def test_xyzw(self):
        v = Vector(1, 2, 3, 4)
        self.assertEqual(v[0], 1)
        self.assertEqual(v[1], 2)
        self.assertEqual(v[2], 3)
        self.assertEqual(v[3], 4)
        self.assertEqual(v.x, 1)
        self.assertEqual(v.y, 2)
        self.assertEqual(v.z, 3)
        self.assertEqual(v.w, 4)
        self.assertEqual(v.xy, Vector(1, 2))
        v[0] = 11
        v[1] = 12
        v[2] = 13
        v[3] = 14
        self.assertEqual(v[0], 11)
        self.assertEqual(v[1], 12)
        self.assertEqual(v[2], 13)
        self.assertEqual(v[3], 14)
        self.assertEqual(v.x, 11)
        self.assertEqual(v.y, 12)
        self.assertEqual(v.z, 13)
        self.assertEqual(v.w, 14)
        v.x = 101
        v.y = 102
        v.z = 103
        v.w = 104
        self.assertEqual(v[0], 101)
        self.assertEqual(v[1], 102)
        self.assertEqual(v[2], 103)
        self.assertEqual(v[3], 104)
        self.assertEqual(v.x, 101)
        self.assertEqual(v.y, 102)
        self.assertEqual(v.z, 103)
        self.assertEqual(v.w, 104)
        v.xy = Vector(1001, 1002)
        self.assertEqual(v[0], 1001)
        self.assertEqual(v[1], 1002)
        self.assertEqual(v[2], 103)
        self.assertEqual(v[3], 104)
        self.assertEqual(v.x, 1001)
        self.assertEqual(v.y, 1002)
        self.assertEqual(v.z, 103)
        self.assertEqual(v.w, 104)

    def test_add(self):
        v1 = Vector(2, 3, 5)
        v2 = Vector(7, 11, 13)
        v3 = v1 + v2
        self.assertEqual(v3.data, [9, 14, 18])

    def test_sub(self):
        v1 = Vector(2, 3, 5)
        v2 = Vector(7, 11, 13)
        v3 = v1 - v2
        self.assertEqual(v3.data, [-5, -8, -8])

    def test_eq(self):
        self.assertEqual(Vector(), ())
        self.assertEqual(Vector(), [])
        self.assertEqual(Vector(1), Vector(1))
        self.assertEqual(Vector(1, 2), Vector(1, 2))
        self.assertEqual(Vector(1, 2, 3), Vector(1, 2, 3))
        self.assertNotEqual(Vector(1, 2, 3), Vector(4, 2, 3))
        self.assertNotEqual(Vector(1, 2, 3), Vector(1, 4, 3))
        self.assertNotEqual(Vector(1, 2, 3), Vector(1, 2, 4))
        self.assertNotEqual(Vector(1, 2, 3), Vector(1, 2))

    def test_mult(self):
        self.assertEqual(Vector() * 10, ())
        self.assertEqual(Vector(1) * 10, Vector(10))
        self.assertEqual(Vector(1, 2, 3) * 10, Vector(10, 20, 30))
        self.assertEqual(
            Vector(1.0, 2.0, 3.0) / 10,
            Vector(1.0 * (1.0 / 10), 2.0 * (1.0 / 10), 3.0 * (1.0 / 10)),
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2022 Ledger SAS,
# written by Olivier Hériveaux, Manuel San Pedro and Michaël Mouchous


import serial.serialutil
import time
from .vector import Vector
from .exceptions import ProtocolError, ConnectionFailure
from enum import Enum, Flag
from typing import Optional, List, Union
from .stage import Stage


class Link:
    """
    Class to control Newport SMC100 controllers. This uses a serial device to
    communicate with multiple controllers which may be configured in daisy-chain
    configuration.

    Current design allows multiple instance of SMC100 to share the same
    communication channel without being considered as the same stage. For
    example, we can have two XY stages using four SMC100 controllers connected
    in daisy-chain configuration.

    As specified in SMC100 controllers documentation, the address of the first
    controller in the daisy chain is always zero.
    """

    def __init__(self, dev: Optional[str] = None):
        """
        :param dev: Serial device. For instance `'/dev/ttyUSB0'`.
            If not provided, a suitable device is searched according to
            according to vendor and product IDs
        """
        try:
            self.serial = serial.Serial(port=dev, baudrate=57600, xonxoff=True)
        except serial.serialutil.SerialException as e:
            raise ConnectionFailure() from e

    def send(self, address: Optional[int], command: str):
        """
        Send a command to a controller.

        :param address: Controller address. If None, only the command is sent without prefix.
        :param command: Command string. Don't include address nor CR LF since
            they are added automatically by this method.
        """
        to_send = f'{"" if address is None else address}{command}\r\n'
        self.serial.write(to_send.encode())

    def receive(self) -> Optional[str]:
        """
        Read input serial buffer to get a response. Blocks until a response is
        available.

        :return: Received response string, CR-LF removed.
        """
        # Read at least 2 bytes for CR-LF.
        response = ""
        while True:
            # Communication contains only ASCII characters, for robustness, HSBit is masked
            c = self.serial.read(1)[0] & 0x7F
            if c == ord("\n"):
                return response
            # We may receive null characters after controller reset. Just
            # ignore it, and we will be fine, they are useless anyway.
            elif c not in (ord("\r"), 0):
                response += chr(c)

    def query(
        self, address: Optional[int], command: str, lazy_res: bool = False
    ) -> Optional[str]:
        """
        Send a query.

        :param address: Controller address. int.
        :param command: Command string, without '?'.
        :param lazy_res: If True, response is not fetched. The caller must
            ensure the response will be fetched later. This flag can be used
            for optimization to send a batch of queries and then fetch all the
            response.
        :return: Received response, or None if lazy_res is True.
        """
        self.send(address, command + "?")
        if not lazy_res:
            try:
                return self.response(address, command)
            except ProtocolError:
                # Retry to send the query
                return self.query(address, command, lazy_res)

    def response(self, address: Optional[int], command: str) -> str:
        """
        Get and return the response of a query. Parameters are required to
        check the header of the response.

        :param address: Controller address. int.
        :param command: Command string, without '?'.
        """
        query_string = f'{"" if address is None else address}{command}'
        res = self.receive()
        if res is None or res[: len(query_string)] != query_string:
            raise ProtocolError(query_string, res)
        return res[len(query_string) :]


class State(int, Enum):
    """
    Possible controller states.
    The values in this enumeration corresponds to the values returned by the
    controller.
    """

    NOT_REFERENCED_FROM_RESET = 0x0A
    NOT_REFERENCED_FROM_HOMING = 0x0B
    NOT_REFERENCED_FROM_CONFIGURATION = 0x0C
    NOT_REFERENCED_FROM_DISABLE = 0x0D
    NOT_REFERENCED_FROM_READY = 0x0E
    NOT_REFERENCED_FROM_MOVING = 0x0F
    NOT_REFERENCED_STAGE_ERROR = 0x10
    NOT_REFERENCED_FROM_JOGGING = 0x11
    CONFIGURATION = 0x14
    HOMING_RS232 = 0x1E
    HOMING_SMCRC = 0x1F
    MOVING = 0x28
    READY_FROM_HOMING = 0x32
    READY_FROM_MOVING = 0x33
    READY_FROM_DISABLE = 0x34
    READY_FROM_JOGGING = 0x35
    DISABLE_FROM_READY = 0x3C
    DISABLE_FROM_MOVING = 0x3D
    DISABLE_FROM_JOGGING = 0x3E
    JOGGING_FROM_READY = 0x46
    JOGGING_FROM_DISABLE = 0x47


class Error(int, Flag):
    """
    Information returned when querying positioner error.
    """

    OUTPUT_POWER_EXCEEDED = 1 << 9
    DC_VOLTAGE_TOO_LOW = 1 << 8
    WRONG_STAGE = 1 << 7
    HOMING_TIMEOUT = 1 << 6
    FOLLOWING_ERROR = 1 << 5
    SHORT_CIRCUIT = 1 << 4
    RMS_CURRENT_LIMIT = 1 << 3
    PEAK_CURRENT_LIMIT = 1 << 2
    POSITIVE_END_OF_RUN = 1 << 1
    NEGATIVE_END_OF_RUN = 1 << 0
    NO_ERROR = 0


class ErrorAndState:
    """
    Information returned when querying positioner error and controller state.
    """

    state = State.NOT_REFERENCED_FROM_RESET
    error = Error.NO_ERROR

    @property
    def is_referenced(self) -> bool:
        """
        :return: True if state is not one of the NOT_REFERENCED_x states.
        """
        if self.state is None:
            raise RuntimeError("state not available")
        else:
            return not (
                (self.state >= State.NOT_REFERENCED_FROM_RESET)
                and (self.state <= State.NOT_REFERENCED_FROM_JOGGING)
            )

    @property
    def is_ready(self) -> bool:
        """:return: True if state is one of READY_x states."""
        return (self.state >= State.READY_FROM_HOMING) and (
            self.state <= State.READY_FROM_JOGGING
        )

    @property
    def is_moving(self) -> bool:
        """:return: True if state is MOVING."""
        return self.state == State.MOVING

    @property
    def is_homing(self) -> bool:
        """:return: True if state is one of HOMING_x states."""
        return (self.state >= State.HOMING_RS232) and (self.state <= State.HOMING_SMCRC)

    @property
    def is_jogging(self) -> bool:
        """:return: True if state is one of JOGGING_x states."""
        return (self.state >= State.JOGGING_FROM_READY) and (
            self.state <= State.JOGGING_FROM_DISABLE
        )

    @property
    def is_disabled(self) -> bool:
        """:return: True if state is one of DISABLE_x states."""
        return (self.state >= State.DISABLE_FROM_READY) and (
            self.state <= State.DISABLE_FROM_JOGGING
        )


class SMC100(Stage):
    """
    Class to command Newport SMC100 controllers.
    """

    def __init__(self, dev: Optional[Union[str, Link, "SMC100"]], addresses: List[int]):
        """
        :param dev: Serial device string (for instance `'/dev/ttyUSB0'` or
            'COM0'), an instance of Link, or an instance of SMC100 sharing
            the same serial device.
            If not provided, a suitable device is searched according to
            according to vendor and product IDs
        :param addresses: An iterable of int controller addresses.
        """
        super().__init__(num_axis=len(addresses))
        if isinstance(dev, Link):
            self.link = dev
        elif isinstance(dev, SMC100):
            self.link = dev.link
        else:
            self.link = Link(dev)
        self.addresses = addresses
        # Hack: sometimes the controllers may be in error stage when
        # initialized (for instance if the stage is not the same as the
        # memorized one).
        # Errors prevent any command being executed. Reading the controller
        # state at startup returns and clear error flags.
        for addr in self.addresses:
            self.link.query(addr, "TS")

    @property
    def position(self):
        """
        Stage position, in micrometers.

        :getter: Query and return current stage position.
        :setter: Move stage.
        """
        result = Vector(dim=self.num_axis)
        # It is faster to send all the request and then get all the responses.
        # This reduces a lot the latency.
        for i, addr in enumerate(self.addresses):
            res = self.link.query(addr, "TP", lazy_res=False)
            if res is None:
                raise ProtocolError("TP")
            val = float(res)
            result[i] = val
        return result

    @position.setter
    def position(self, value: Vector):
        # To check dimension and range of the given value
        pos_setter = Stage.position.fset
        assert pos_setter is not None
        pos_setter(self, value)

        # Enable the motors
        self.is_disabled = False
        commands = []
        for position, addr in zip(value.data, self.addresses):
            commands.append(f"{addr}PA{position:.5f}")
        self.link.send(None, "\r\n".join(commands))

    def home(self, wait=False):
        """
        Perform home search.

        :param wait: Optionally waits for move operation to be done.
        """
        self.home_search()
        if wait:
            self.wait_move_finished()

    def home_search(self):
        """
        Perform home search.
        Home search is performed even if the axes are already referenced.
        It may be better to use home_search_if_required.
        """
        for addr in self.addresses:
            # It has been observed that OR may be not effective if another command (like TS?) is not
            # performed before
            self.get_error_and_state(addr=addr)
            self.link.send(addr, "OR")

    def home_search_if_required(self):
        """
        Perform home search for all axes which are not referenced.
        """
        for addr in self.addresses:
            state = self.get_error_and_state(addr=addr)
            if not state.is_referenced:
                self.link.send(addr, "OR")

    def get_error_and_state(self, addr: int):
        """
        Query current motion controller errors and state.
        Querying the error and state may clear error flags.

        :param addr: Address of the axis.
        :return: Current error and state, in a ErrorAndState instance.
        """
        res = self.link.query(addr, "TS")
        if res is None or len(res) != 6:
            raise ProtocolError("TS", res)
        result = ErrorAndState()
        result.error = Error(int(res[:4], 16))
        result.state = State(int(res[4:], 16))
        return result

    def enter_configuration_state(self, addr: int):
        """Enter configuration state."""
        self.link.send(addr, "PW1")

    def leave_configuration_state(self, addr):
        """
        Leave configuration state. If defined parameters are valid, the
        controller saves them in the flash memory.
        """
        self.link.send(addr, "PW0")

    def controller_address(self, addr: int):
        """
        Get controller's RS-485 address. int in [2, 31].
        """
        res = self.link.query(addr, "SA")
        if res is None:
            raise ProtocolError("SA")
        return int(res)

    def set_controller_address(self, addr: int, value: int):
        """
        Set controller's RS-485 address. int in [2, 31].
        Changing the address is only possible when the controller is in
        configuration state.
        """
        if value not in range(2, 32):
            raise ValueError("Invalid controller address")
        self.link.send(addr, f"SA{value}")

    def move_relative(self, addr: int, offset: float):
        """
        Moves relatively an axis from a given offset

        :param addr: addr of axis to move
        :param offset: offset value
        """
        # Enable the motors
        self.is_disabled = False
        self.link.send(addr, f"PR{offset:.5f}")

    def stop(self, addr: Optional[int] = None):
        """
        Stops the motion on an axis. On all axis if addr not specified.

        :param addr: Address of the axis to stop. If None, stop all the controllers
        """
        self.link.send(addr, "ST")

    def reset(self, addr: Optional[int] = None):
        """
        Resets the controller at specified address. For all controllers if not specified

        :param addr: address of the controller to reset
        """
        for addr in self.addresses if addr is None else [addr]:
            self.link.send(addr, "RS")

    def set_position(self, addr: int, value: float, blocking=True):
        """
        Sets the position of a single axis

        :param addr: address of the axis to set the position
        :param value: stage position, in micrometers
        :param blocking: if True, blocking mode: wait for the position to be
            reached before exit.
        """
        # Enable the motors
        self.is_disabled = False
        self.link.send(addr, f"PA{value:.5f}")

        if blocking:
            error_and_state = self.get_error_and_state(addr=addr)
            while error_and_state.is_moving:
                time.sleep(0.1)

    @property
    def is_moving(self) -> bool:
        """
        Indicates if the stage is currently moving due to MOVE, HOME or JOG operation.

        :return: Moving state of the stage
        """
        for addr in self.addresses:
            state = self.get_error_and_state(addr=addr)
            if state.is_moving or state.is_homing or state.is_jogging:
                return True
        return False

    @property
    def is_disabled(self) -> bool:
        """
        Indicates if the stage is currently in DISABLE state.
        :return: Disabled state of the stage
        """
        for addr in self.addresses:
            state = self.get_error_and_state(addr=addr)
            if state.is_disabled:
                return True
        return False

    @is_disabled.setter
    def is_disabled(self, value: bool):
        # self.is_disabled = True makes enter DISABLE state
        self.enter_leave_disable_state(None, enter=value)

    def enter_leave_disable_state(self, addr: Optional[int], enter: bool = True):
        """
        Permits for a specified axis to enter or leave the DISABLE state.
        DISABLE state makes the motor not energized and opens the control loop.

        :param addr: address of the axis to operate.
            If None is passed, it applies to all controllers
        :param enter: True to enter, False to leave DISABLE state
        """
        # MM0 changes the controller’s state from READY to DISABLE (enter)
        # MM1 changes the controller’s state from DISABLE to READY (leave)
        self.link.send(addr, f"MM{0 if enter else 1}")



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Values and descriptions of commands and error codes has been taken from GRBL Github repository
# https://github.com/grbl/grbl
# and the instruction manual of the CNC 3018 PRO
# https://drive.google.com/file/d/1yQH9gtO8lWbE-K0dff8g9zq_1xOB57x7
#
# Copyright 2018-2022 Ledger SAS, written by Michaël Mouchous
from enum import Enum, Flag
from typing import Union


class InvertMask(Flag):
    """
    Invert Mask flags values
    """

    INVERT_X = 1
    INVERT_Y = 1 << 1
    INVERT_Z = 1 << 2


class StatusReportMask(Flag):
    """
    Status report flags values
    """

    NOTHING = 0
    MACHINE_POSITION = 1
    WORK_POSITION = 1 << 1
    PLANNER_BUFFER = 1 << 2
    RX_BUFFER = 1 << 3
    LIMIT_PINS = 1 << 4


class GRBLSetting(Enum):
    """
    GRBL Setting are obtained by sending the '$$' command as a list of
    key-value pairs '$K=V' with K being a number.
    """

    STEP_PULSE = "$0"
    STEP_IDLE_DELAY = "$1"
    STEP_PORT_INVERT = "$2"
    DIR_PORT_INVERT = "$3"
    STEP_ENABLE_INVERT = "$4"
    LIMIT_PINS_INVERT = "$5"
    PROBE_PIN_INVERT = "$6"
    STATUS_REPORT_MASK = "$10"
    JUNCTION_DEVIATION = "$11"
    ARC_TOLERANCE = "$12"
    REPORT_INCHES = "$13"
    SOFT_LIMITS = "$20"
    HARD_LIMITS = "$21"
    HOMING_CYCLE = "$22"
    HOMING_DIR_INVERT = "$23"
    HOMING_FEED = "$24"
    HOMING_SEEK = "$25"
    HOMING_DEBOUNCE = "$26"
    HOMING_PULL_OFF = "$27"
    SPINDLE_RPM_MAX = "$30"
    SPINDLE_RPM_MIN = "$31"
    LASER_MODE = "$32"
    STEPS_PER_MM_X = "$100"
    STEPS_PER_MM_Y = "$101"
    STEPS_PER_MM_Z = "$102"
    MAX_RATE_X = "$110"
    MAX_RATE_Y = "$111"
    MAX_RATE_Z = "$112"
    ACCELERATION_X = "$120"
    ACCELERATION_Y = "$121"
    ACCELERATION_Z = "$122"
    MAX_TRAVEL_X = "$130"
    MAX_TRAVEL_Y = "$131"
    MAX_TRAVEL_Z = "$132"

    @property
    def type(self) -> type:
        """
        Gives the type of the value stored in the GRBL setting
        """
        return type(self._description[0])

    @property
    def default_value(self) -> Union[float, bool, InvertMask, StatusReportMask, int]:
        """
        Gives the default value of the GRBL setting
        """
        return self._description[0]

    @property
    def description(self) -> str:
        """
        Gives the string description of the GRBL setting
        """
        return self._description[1]

    @property
    def _description(self) -> dict:
        # Numbering: (type, default value, description)
        return {
            GRBLSetting.STEP_PULSE: (10.0, "Step pulse, usec"),
            GRBLSetting.STEP_IDLE_DELAY: (25.0, "Step idle delay, msec"),
            GRBLSetting.STEP_PORT_INVERT: (InvertMask(0), "Step port invert"),
            GRBLSetting.DIR_PORT_INVERT: (InvertMask(6), "Direction port invert"),
            GRBLSetting.STEP_ENABLE_INVERT: (False, "Step enable invert"),
            GRBLSetting.LIMIT_PINS_INVERT: (False, "Limit pins invert"),
            GRBLSetting.PROBE_PIN_INVERT: (False, "Probe pin invert"),
            GRBLSetting.STATUS_REPORT_MASK: (StatusReportMask(3), "Status report mask"),
            GRBLSetting.JUNCTION_DEVIATION: (0.020, "Junction deviation, mm"),
            GRBLSetting.ARC_TOLERANCE: (0.002, "Arc tolerance, mm"),
            GRBLSetting.REPORT_INCHES: (False, "Report inches"),
            GRBLSetting.SOFT_LIMITS: (False, "Soft limits"),
            GRBLSetting.HARD_LIMITS: (False, "Hard limits"),
            GRBLSetting.HOMING_CYCLE: (False, "Homing cycle"),
            GRBLSetting.HOMING_DIR_INVERT: (InvertMask(1), "Homing dir invert"),
            GRBLSetting.HOMING_FEED: (50.000, "Homing feed, mm/min"),
            GRBLSetting.HOMING_SEEK: (635.000, "Homing seek, mm/min"),
            GRBLSetting.HOMING_DEBOUNCE: (250.0, "Homing debounce, msec"),
            GRBLSetting.HOMING_PULL_OFF: (1.000, "Homing pull-off, mm"),
            GRBLSetting.STEPS_PER_MM_X: (800.0, "X steps/mm"),
            GRBLSetting.STEPS_PER_MM_Y: (800.0, "Y steps/mm"),
            GRBLSetting.STEPS_PER_MM_Z: (800.0, "Z steps/mm"),
            GRBLSetting.MAX_RATE_X: (635.000, "X max rate, mm/min"),
            GRBLSetting.MAX_RATE_Y: (635.000, "Y max rate, mm/min"),
            GRBLSetting.MAX_RATE_Z: (635.000, "Z max rate, mm/min"),
            GRBLSetting.ACCELERATION_X: (50.000, "X acceleration, mm/sec^2"),
            GRBLSetting.ACCELERATION_Y: (50.000, "Y acceleration, mm/sec^2"),
            GRBLSetting.ACCELERATION_Z: (50.000, "Z acceleration, mm/sec^2"),
            GRBLSetting.MAX_TRAVEL_X: (225.000, "X max travel, mm"),
            GRBLSetting.MAX_TRAVEL_Y: (125.000, "Y max travel, mm"),
            GRBLSetting.MAX_TRAVEL_Z: (170.000, "Z max travel, mm"),
            GRBLSetting.SPINDLE_RPM_MAX: (
                1000.0,
                "Spindle maximal rotation speed, rpm",
            ),
            GRBLSetting.SPINDLE_RPM_MIN: (0.0, "Spindle minimal rotation speed, rpm"),
            GRBLSetting.LASER_MODE: (False, "Laser mode activated"),
        }[self]

    def __str__(self):
        return f"{self.description}: (default value: {self.type(self.default_value)})"



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018 Ledger SAS, written by Olivier Hériveaux

from .stage import Stage
from .corvus import Corvus
from .m3fs import M3FS
from .smc100 import SMC100
from .autofocus import Autofocus
from .vector import Vector
from .tic import Tic, TicDirection
from .cncrouter import CNCRouter, CNCStatus

__all__ = [
    "Stage",
    "Corvus",
    "M3FS",
    "SMC100",
    "Autofocus",
    "Vector",
    "Tic",
    "TicDirection",
    "CNCRouter",
    "CNCStatus",
]



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2022 Ledger SAS, written by Michaël Mouchous

from .vector import Vector
from typing import Optional, Callable
from abc import ABC, abstractmethod
from serial.tools.list_ports import comports


class Stage(ABC):
    """
    Stage is an abstract class from which all stage implementation must inherit to get a consistent
    behavior and value checking for getting and setting position, setting software limits...
    """

    def __init__(self, num_axis=1):
        """
        :param num_axis: The number of axis of the stage, can be updated or set after initialization
            of the object.
        """
        self.num_axis = num_axis
        # The wait routine is a function that is called when the wait_move_finished is looping.
        # It can be used to add some temporization and/or UI updates.
        self.wait_routine: Optional[Callable] = None

        # Minimum and maximum software limits
        self._minimums: Optional[Vector] = None
        self._maximums: Optional[Vector] = None

    def find_device(
        self,
        pid: Optional[int] = None,
        vid: Optional[int] = None,
        serial_number: Optional[str] = None,
    ) -> str:
        """
        Find automatically one device according to given information in parameters

        :param pid: Product ID of the device. If None, not a limiting criteria.
        :param vid: Vendor ID of the device. If None, not a limiting criterial.
        :param serial_number: serial number of the device. If not None, the device
            must have a serial number and matches the given value.
        :return: The device path to the communication port.
        """
        possible_ports = []
        for port in comports():
            if serial_number is not None:
                if port.serial_number == serial_number:
                    possible_ports.append(port)
                continue
            if (port.pid, port.vid) == (pid, vid):
                possible_ports.append(port)
        if len(possible_ports) > 1:
            raise RuntimeError("Multiple devices found! I don't know which one to use.")
        elif len(possible_ports) == 1:
            dev = possible_ports[0].device
        else:
            raise RuntimeError("No device found")
        return dev

    @property
    @abstractmethod
    def position(self) -> Vector:
        """
        Current stage position. Vector.

        :getter: Query and return stage position.
        :setter: Move the stage.
        """
        ...

    @position.setter
    def position(self, value: Vector):
        """
        Set stage to a new destination.
        """
        # Checks dimension according to number of axis, and range according to minimums/maximums
        # (raises an exception if incorrect)
        self.check_dimension(value)
        self.check_range(value)

    @property
    @abstractmethod
    def is_moving(self) -> bool:
        """Queries the status of the stage to determine if it is moving"""
        ...

    def wait_move_finished(self):
        """
        Wait until all axis are ready.
        """
        while self.is_moving:
            if self.wait_routine is not None:
                self.wait_routine()

    def move_to(self, value: Vector, wait: bool = True):
        """
        Move stage to a new position, with the possibility to poll device until the move finishes.

        :param value: New stage position.
        :param wait: If True, wait for stage to be at the new position,
            otherwise return immediately.
        """
        self.position = value
        if wait:
            self.wait_move_finished()

    def check_dimension(self, value: Vector):
        """
        Check if the given value's dimension is consistent according to the number of axes of the
        Stage.
        :param value: The vector to check
        :raises ValueError: If dimension is inconsistent with the number of axis of the Stage.
        """
        if self.num_axis != len(value):
            raise ValueError(
                f"The given vector's dimension {len(value)} is incorrect: expected {self.num_axis}"
            )

    def check_range(self, value: Vector):
        """
        Check if the given value is in range according to minimums/maximums.
        If it is not the case, a ValueException is raised.

        :param value: The vector to check if it is within authorized range.
        """
        if self._minimums is None and self._maximums is None:
            # No restriction, we can return
            return
        # Get minimums and/or maximums. If one of them is None, we use the value.
        mins = value if self._minimums is None else self._minimums
        maxs = value if self._maximums is None else self._maximums

        # Check if the dimensions are correct.
        if len(mins) != len(value) or len(maxs) != len(value):
            raise ValueError(
                "Invalid vector dimension according to minimums and/or maximums"
            )

        for i in range(len(value)):
            if mins[i] is not None and value[i] < mins[i]:
                raise ValueError(
                    f"Invalid value for dimension {i}: {value[i]} "
                    f"is smaller that minimum {mins[i]}."
                )
            if maxs[i] is not None and value[i] > maxs[i]:
                raise ValueError(
                    f"Invalid value for dimension {i}: {value[i]} "
                    f"is greater that maximum {maxs[i]}."
                )

    @property
    def minimums(self) -> Optional[Vector]:
        return self._minimums

    @minimums.setter
    def minimums(self, value: Optional[Vector]):
        if value is not None:
            self.check_dimension(value)
        self._minimums = value

    @property
    def maximums(self) -> Optional[Vector]:
        return self._maximums

    @maximums.setter
    def maximums(self, value: Optional[Vector]):
        if value is not None:
            self.check_dimension(value)
        self._maximums = value

    def home(self, wait=False):
        """Triggers a homing command.

        :param wait: Optionally waits for move operation to be done.
        """
        self.move_to(Vector(dim=self.num_axis), wait=wait)



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Values and descriptions of commands and error codes has been taken from GRBL Github repository
# https://github.com/grbl/grbl
# and the instruction manual of the CNC 3018 PRO
# https://drive.google.com/file/d/1yQH9gtO8lWbE-K0dff8g9zq_1xOB57x7
#
# Copyright 2018-2023 Ledger SAS, written by Michaël Mouchous

import serial.serialutil
import time
from typing import Optional, Tuple, List, Union
from .exceptions import ConnectionFailure
from .vector import Vector
from enum import Enum
from .grbl import GRBLSetting, InvertMask, StatusReportMask
from .stage import Stage


class CNCStatus(str, Enum):
    """
    Possible statuses that the CNC can report.
    """

    IDLE = "Idle"
    RUN = "Run"
    HOLD = "Hold"
    DOOR = "Door"
    HOME = "Home"
    ALARM = "Alarm"
    CHECK = "Check"


class CNCError(Exception):
    """Exception raised when a specific error is detected by the CNC"""

    def __init__(self, message: str, cncstatus: CNCStatus):
        super().__init__(message)
        self.cncstatus = cncstatus


class CNCRouter(Stage):
    """
    Class to command CNC routers.
    """

    def __init__(self, dev: Optional[str] = None, reset_wait_time=2.0):
        """
        Open serial device to connect to the CNC routers. Raise a
        ConnectionFailure exception if the serial device could not be open.

        :param dev: Serial device. For instance `'/dev/ttyUSB0'`. If not provided, try
            to discover a suitable device according to device vendor and product IDs
        :param reset_wait_time: Depending on the state of the stage, it can take some time for
            GRBL to reset. This parameter makes the wait time to be tuned, by giving a time in seconds.
        """

        super().__init__(num_axis=3)
        self.reset_wait_time = reset_wait_time
        try:
            dev = dev or self.find_device(pid=0x7523, vid=0x1A86)
            self.serial = serial.Serial(dev, 115200, timeout=1)
        except serial.serialutil.SerialException as e:
            raise ConnectionFailure() from e
        self.reset_grbl()

    def reset_grbl(self, wait_time: Optional[float] = None) -> bool:
        """
        Sends a CTRL+X control to reset the GRBL

        :return: True if the GRBL sent the correct prompt at the end of the reset
        :param wait_time: Depending on the state of the stage, it can take some time for GRBL to
            reset. This parameter makes the wait time to be tuned, by giving a time in seconds.
        """
        if wait_time is None:
            wait_time = self.reset_wait_time
        self.serial.flush()
        self.send("\030", eol="")
        time.sleep(wait_time)
        self.send("")

        responses = self.receive_lines()
        # Expected ATR is 'Grbl x.xx ['$' for help]' for first line.
        ok = responses[0].startswith("Grbl") and responses[0].endswith("['$' for help]")
        if not ok:
            return False
        # Unlock if necessary
        if "[MSG:'$H'|'$X' to unlock]" in responses:
            ok = self.unlock()
        # Force status report to send Machine Position and Work Position
        self.set_grbl_setting(
            GRBLSetting.STATUS_REPORT_MASK,
            StatusReportMask.MACHINE_POSITION | StatusReportMask.WORK_POSITION,
        )
        return ok

    def sleep(self) -> str:
        """
        Sends a `$SLP` command. The stage responds a message `[MSG:Sleeping]` after `ok`.

        :return: The response of the CNC ('ok' if command has been submitted correctly).
        """
        self.send_receive("$SLP")
        return self.receive()

    def unlock(self) -> bool:
        """
        Unlock the motor. It may happen when the stage has gone further its limits,
        and raised an alarm, or has been disabled when going in sleep mode (`$SLP`)

        :return: `True` if message `[MSG:Caution: Unlock]` has been returned
        """
        self.send("$X")
        return "[MSG:Caution: Unlocked]" in self.receive_lines()

    def get_grbl_setting(
        self, setting: GRBLSetting
    ) -> Union[float, bool, InvertMask, StatusReportMask, int]:
        return self.get_grbl_settings()[setting]

    def set_grbl_setting(
        self,
        setting: GRBLSetting,
        value: Union[float, bool, InvertMask, StatusReportMask, int],
    ):
        """
        Set the GRBL setting of the Router with given value. The value type must correspond to
        type defined in :py:class:`GRBLSetting`.
        """
        if setting.type != type(value):
            ValueError(f"The setting {setting} expects a value of type {setting.type}.")
        if isinstance(value, (InvertMask, StatusReportMask)):
            # Get the int value of the flag
            value = value.value
        elif isinstance(value, bool):
            # Get the int value of the boolean (1 if true, 0 if false)
            value = 1 if value else 0
        # We do nothing for int and floats.
        return self.send_receive(f"{setting.value}={value}")

    def get_grbl_settings(self) -> dict:
        """
        Obtains and parse the list of GRBLSettings with the `$$` command.

        :return: A dictionary containing the GRBLSetting as key and its corresponding value
        """
        self.send("$$")
        lines = self.receive_lines()
        settings = {}
        for line in lines:
            key, value = line.split("=", 1)
            setting = GRBLSetting(key)
            if setting.type == float:
                value = float(value)
                settings[setting] = value
            else:
                # For (bool, InvertMask, StatusReportMask, int) types, str must be changed to int
                # first.
                settings[setting] = setting.type(int(value))
        return settings

    def get_current_status(self) -> Optional[Tuple[CNCStatus, dict]]:
        """
        Sending '?' character permits to get the status of the CNC
        router

        :return: A tuple containing the status and a dictionary of all other parameters in the
            output of the command.
        """

        self.send("?", eol="")
        status = self.receive()

        # Retry, sometimes it does not respond
        if status == "":
            self.send("?", eol="")
            status = self.receive()

        # Sometimes, the CNC returns 'ok' and the actual response is following.
        while status == "ok":
            status = self.receive()

        # In the case there has been a communication error
        if status is None:
            return None

        # The possible outputs
        # '<Idle|MPos:1.000,3.000,4.000|FS:0,0|WCO:0.000,0.000,0.000>'
        # 'ALARM:1'

        if status.startswith("ALARM:1"):
            # The ALARM message is followed by something like
            # '[MSG:Reset to continue]'
            next = self.receive()
            raise CNCError(next, CNCStatus.ALARM)

        # Discard any unwanted format
        if not (status.startswith("<") and status.endswith(">")):
            print(f"Response to '?' is unexpected: {status}")
            return None

        # Remove the chevrons and split all pipes.
        elements = status[1:-1].split("|")

        # First element is the CNC status
        cncstatus = CNCStatus(elements[0])

        # Next elements to be parsed as key/value pairs
        others = {}
        for key_value in [element.split(":", 1) for element in elements[1:]]:
            key = key_value[0]
            value = None if len(key_value) == 1 else key_value[1]
            if value is not None and "," in value:
                value = value.split(",")

            others[key] = value

        return cncstatus, others

    def send(self, command: str, eol: Optional[str] = None):
        """
        Send a command.

        :param command: Command string.
        :param eol: End of command character(s). If None, LF will be automatically append.
        """
        if eol is None:
            eol = "\n"
        self.serial.write((command + eol).encode())

    def receive_lines(self, until: str = "ok") -> List[str]:
        """
        Receive multiple lines until getting as specific value.

        :param until: The expected response indicating the end of received lines.
        :return: The list of all received lines. Note that the expected line is not included in the
            list.
        """
        lines = []
        while (line := self.serial.readline().strip().decode()) != until:
            if len(line):
                lines.append(line)
        return lines

    def receive(self) -> str:
        """
        Read input serial buffer to get a response. Blocks until a response is
        available.

        :return: Received response string, CR-LF removed.
        """
        tries = 10
        # Read at least 2 bytes for CR-LF.
        response = self.serial.read(2)
        while response[-2:] != b"\r\n":
            part = self.serial.read(1)
            response += part
            # Give a chance to get out of this
            if part == b"":
                tries -= 1
            if tries == 0:
                return ""
        # Remove CR-LF and return as string
        return response[:-2].decode()

    def send_receive(self, command: str) -> str:
        """
        Send a command, wait and return the response.

        :param command: Command string.
        :return: Received response string, CR-LF removed.
        """
        self.send(command)
        return self.receive()

    @property
    def position(self) -> Vector:
        """
        Current stage position. Vector.

        :getter: Query and return stage position.
        :setter: Move the stage.
        """
        extra_dict = {}
        # We loop until we get a current status providing 'WCO' which stores the
        # origin.
        while "WCO" not in extra_dict.keys():
            current_status = self.get_current_status()
            extra_dict = current_status[1] if current_status is not None else {}

        mpos = Vector(*tuple(float(x) for x in extra_dict["MPos"]))
        wco = Vector(*tuple(float(x) for x in extra_dict["WCO"]))

        return mpos - wco

    @position.setter
    def position(self, value: Vector):
        # To check dimension and range of the given value
        pos_setter = Stage.position.fset
        assert pos_setter is not None
        pos_setter(self, value)

        command = f"G0 X{value.x}"
        if len(value) > 1:
            command += f" Y{value.y}"
        if len(value) > 2:
            command += f" Z{value.z}"
        self.send_receive(command)

    @property
    def is_moving(self) -> bool:
        """
        Queries the current status of the CNC in order to determine if the CNC is moving

        :return: True if the CNC reports that a cycle is running (Run) or
            if it is in a middle of a homing cycle (Home).
        """
        while (status := self.get_current_status()) is None:
            pass
        return status[0] in [CNCStatus.RUN, CNCStatus.HOME]

    def set_origin(self) -> str:
        """
        Set current position as origin

        :return: The response of the CNC ('ok' if command has been submitted correctly).
        """
        return self.send_receive("G92 X0 Y0 Z0")

    def home(self, wait=False):
        """
        Sends a `$H` command. The stage responds a message `[MSG:Sleeping]` after `ok`.
        Take caution for collisions before calling this method !

        :param wait: Optionally waits for move operation to be done.
        """
        self.send("$H")
        if wait:
            self.wait_move_finished()



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2022 Ledger SAS, written by Olivier Hériveaux


import serial.serialutil
import time
from .exceptions import ConnectionFailure
from .vector import Vector
from .stage import Stage
from typing import Optional


class Corvus(Stage):
    """
    Class to command Corvus Eco XYZ stage controller.
    """

    def __init__(self, dev: Optional[str] = None, serial_number: Optional[str] = None):
        """
        Open serial device to connect to the Corvus controller. Raise a
        ConnectionFailure exception if the serial device could not be open.

        :param dev: Serial device path. For instance '/dev/ttyUSB0'.
        :param serial_number: Device Serial Number. For instance 'A600AAAA'.
        """
        super().__init__(num_axis=3)
        try:
            dev = dev or self.find_device(serial_number=serial_number)
            self.serial = serial.Serial(dev, 57600)
        except serial.serialutil.SerialException as e:
            raise ConnectionFailure() from e
        self.__init()

    def __init(self):
        """Initialize a few parameters."""
        # Tell we work with 3 axis for move commands
        self.send("3 setdim")
        # Set unit to micrometers, for all axis
        unit = 1
        self.send("{0} -1 setunit".format(unit))
        # Verify it has been taken into account (this can be dangerous
        # otherwise)
        res = self.send_receive("-1 getunit").split()
        for x in res:
            assert int(x) == unit
        # Enable joystick
        self.enable_joystick()

    def send(self, command: str):
        """
        Send a command.

        :param command: Command string. CR or blank must not be present at the
            end of this string.
        """
        self.serial.write((command + " ").encode())

    def receive(self) -> str:
        """
        Read input serial buffer to get a response. Blocks until a response is
        available.

        :return: Received response string, CR-LF removed.
        """
        # Read at least 2 bytes for CR-LF.
        response = self.serial.read(2)
        while response[-2:] != b"\r\n":
            response += self.serial.read(1)
        # Remove CR-LF and return as string
        return response[:-2].decode()

    def send_receive(self, command: str) -> str:
        """
        Send a command, wait and return the response.

        :param command: Command string. CR or blank must not be present at the
            end of this string.
        :return: Received response string, CR-LF removed.
        """
        self.send(command)
        return self.receive()

    def calibrate_xy(self):
        """
        Execute limit-switch move only on X and Y axises. Wait until
        calibration si done.
        Take caution for collisions before calling this method !
        """
        # Disable Z-axis
        self.send("3 3 setaxis")
        # Call for calibration
        self.send("cal")
        self.send("rm")
        # Wait until finished
        for i in range(2):
            while int(self.send_receive("{0} getcaldone".format(i + 1))) != 3:
                time.sleep(0.1)
        # Re-enabled Z-axis
        self.send("1 3 setaxis")

    def calibrate(self):
        """
        Execute limit-switch move. Wait until calibration is done.
        Take caution for collisions before calling this method !
        """
        # Call for calibration
        self.send("cal")
        self.send("rm")
        # Wait until finished
        for i in range(3):
            while int(self.send_receive("{0} getcaldone".format(i + 1))) != 3:
                time.sleep(0.1)

    def home(self, wait=False):
        """
        Execute limit-switch move.
        Take caution for collisions before calling this method !

        :param wait: Optionally waits for move operation to be done.
        """
        # Call for calibration
        self.send("cal")
        if wait:
            self.wait_move_finished()

    def move_relative(self, x, y, z):
        """
        Move stage relative to current position.

        :param x: Relative distance on X-axis, in micrometers.
        :param y: Relative distance on Y-axis, in micrometers.
        :param z: Relative distance on Z-axis, in micrometers.
        """
        self.send("3 setdim")
        self.send("{0} {1} {2} rmove".format(x, y, z))
        self.wait_move_finished()

    @property
    def is_moving(self):
        return bool(int(self.send_receive("st")) & 1)

    def set_origin(self):
        """Set current stage's coordinates as the new origin."""
        self.send("0 0 0 setpos")

    def enable_joystick(self):
        """Enable joystick (manual mode)"""
        self.send("1 j")

    @property
    def is_connected(self):
        """:return: True if the instance is connected to the stage."""
        return self.serial is not None

    @property
    def position(self):
        """
        Current stage position. Vector.

        :getter: Query and return stage position.
        :setter: Move the stage.
        """
        res = self.send_receive("p").split()
        return Vector(*tuple(float(x) for x in res))

    @position.setter
    def position(self, value: Vector):
        # To check dimension and range of the given value
        pos_setter = Stage.position.fset
        assert pos_setter is not None
        pos_setter(self, value)

        self.send("3 setdim")
        self.send("{0} {1} {2} move".format(value.x, value.y, value.z))

    @property
    def velocity(self):
        """
        Motors velocity setting, in µm/s.

        :getter: Query and return current setting.
        :setter: Update controller setting.
        """
        res = float(self.send_receive("gv"))
        return res

    @velocity.setter
    def velocity(self, value):
        if value < 0:
            raise ValueError("Velocity parameter cannot be negative.")
        self.send("{0} sv".format(value))

    @property
    def acceleration(self):
        """
        Motors acceleration setting, in µm/s^2.

        :getter: Query and return current setting.
        :setter: Update controller setting.
        """
        res = float(self.send_receive("ga"))
        return res

    @acceleration.setter
    def acceleration(self, value):
        if value < 0:
            raise ValueError("Acceleration parameter cannot be negative.")
        self.send("{0} sa".format(value))

    def save(self):
        """Save current parameters in non-volatile memory."""
        self.send("save")

    def restore(self):
        """
        Reactivates the last saved parameters. Beware this might change units,
        which may be dangerous if care is not taken.
        """
        self.send("restore")



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018 Ledger SAS, written by Olivier Hériveaux


# This file regroups a set of exception classes used by the different stage
# implementations


class ConnectionFailure(Exception):
    pass


class ProtocolError(Exception):
    def __init__(self, query=None, response=None):
        super().__init__(",".join([repr(query), repr(response)]))
        self.query = query
        self.response = response

    def __str__(self):
        return f"ProtocolError({repr(self.query)}, {repr(self.response)})"


class VersionNotSupported(Exception):
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return self.version



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2020 Ledger SAS, written by Olivier Hériveaux

import numpy as np
import unittest


class Autofocus:
    """
    Utility class to add autofocus support for scanning software. Given a set
    of focused coordinates, this class can calculate the correct Z-depth for
    other points.
    """

    def __init__(self):
        self.registered_points = []

    def register(self, x, y, z):
        """
        Register a new focused point.

        :param x: Abscissa of the point.
        :param y: Ordinate of the point.
        :param z: Depth of the point.
        """
        self.registered_points.append((x, y, z))

    def clear(self):
        """Remove all registration points."""
        self.registered_points.clear()

    def focus(self, x, y):
        """
        Guess the correct focus depth given abscissa and ordinate of a new
        point. This method will raise a RuntimeError if not enough points have
        been registered.

        **Warning: Always check for the returned value bounds to prevent stage
        dangerous displacements. Bad points registration or bugs can lead to
        unexpected big displacements.**

        :param x: Abscissa of the point.
        :param y: Ordinate of the point.
        :return: Calculated depth.
        """
        if len(self.registered_points) >= 3:
            return self.__focus_3(x, y)
        else:
            raise RuntimeError("Not enough points registered for autofocus")

    def __focus_3(self, x, y):
        """
        Guess correct focus using 3 points.
        :param x: Abscissa of the point.
        :param y: Ordinate of the point.
        :return: Calculated depth.
        """
        a = np.array(self.registered_points[0])
        b = np.array(self.registered_points[1])
        c = np.array(self.registered_points[2])
        p = np.array((x, y))
        ab = b - a
        ac = c - a
        ap = p - a[:2]
        base = np.matrix([ab[:2], ac[:2]]).T
        ap_in_base = np.array((base.I * np.matrix(ap).T).T)
        z = ab[2] * ap_in_base[0, 0] + ac[2] * ap_in_base[0, 1] + a[2]
        return z

    def __len__(self):
        """:return: Number of registered points."""
        return len(self.registered_points)


class TestAutofocus(unittest.TestCase):
    def test_init(self):
        af = Autofocus()
        self.assertEqual(len(af), 0)

    def test_clear(self):
        af = Autofocus()
        af.register(1, 2, 3)
        self.assertEqual(len(af), 1)
        af.clear()
        self.assertEqual(len(af), 0)

    def test_focus(self):
        af = Autofocus()
        with self.assertRaises(RuntimeError):
            af.focus(3, 3)
        af.register(2, 2, 10)
        with self.assertRaises(RuntimeError):
            af.focus(3, 3)
        af.register(2, 3, 11)
        with self.assertRaises(RuntimeError):
            af.focus(3, 3)
        af.register(3, 2, 11)
        self.assertEqual(af.focus(3, 3), 12)


if __name__ == "__main__":
    unittest.main()



# This file is part of pystages
#
# pystages is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2018-2020 Ledger SAS, written by Olivier Hériveaux


import usb.core
import usb.util
from enum import Enum
from time import sleep
from .stage import Stage
from .vector import Vector
from .exceptions import ConnectionFailure


class TicVariable(Enum):
    """
    Variables which can be read using GET_VARIABLE command.
    https://www.pololu.com/docs/0J71/7
    """

    # Each value is a (offset, size, signed) tuple
    OPERATION_STATE = (0x00, 1, False)
    MISC_FLAGS = (0x01, 1, False)
    ERROR_STATUS = (0x02, 2, False)
    ERRORS_OCCURRED = (0x04, 4, False)
    PLANNING_MODE = (0x09, 1, False)
    TARGET_POSITION = (0x0A, 4, True)
    TARGET_VELOCITY = (0x0E, 4, True)
    STARTING_SPEED = (0x12, 4, False)
    MAX_SPEED = (0x16, 4, False)
    MAX_DECELERATION = (0x1A, 4, False)
    MAX_ACCELERATION = (0x1E, 4, False)
    CURRENT_POSITION = (0x22, 4, True)
    CURRENT_VELOCITY = (0x26, 4, True)
    ACTING_TARGET_POSITION = (0x2A, 4, True)
    TIME_SINCE_LAST_STEP = (0x2E, 4, False)
    DEVICE_RESET = (0x32, 1, False)
    VIN_VOLTAGE = (0x33, 2, False)
    UP_TIME = (0x35, 4, False)
    ENCODER_POSITION = (0x39, 4, True)
    RC_PULSE_WIDTH = (0x3D, 2, False)
    ANALOG_READING_SCL = (0x3F, 2, False)
    ANALOG_READING_SDA = (0x41, 2, False)
    ANALOG_READING_TX = (0x43, 2, False)
    ANALOG_READING_RX = (0x45, 2, False)
    DIGITAL_READINGS = (0x47, 1, False)
    PIN_STATES = (0x48, 1, False)
    STEP_MODE = (0x49, 1, False)
    CURRENT_LIMIT = (0x4A, 1, False)
    DECAY_MODE = (0x4B, 1, False)
    INPUT_STATE = (0x4C, 1, False)
    INPUT_AFTER_AVERAGING = (0x4D, 2, False)
    INPUT_AFTER_HYSTERESIS = (0x4F, 2, False)
    INPUT_AFTER_SCALING = (0x51, 4, True)
    LAST_MOTOR_DRIVER_ERROR = (0x55, 1, False)
    AGC_MODE = (0x56, 1, False)
    AGC_BOTTOM_CURRENT_LIMIT = (0x57, 1, False)
    AGC_CURRENT_BOOST_STEP = (0x58, 1, False)
    AGC_FREQUENCY_LIMIT = (0x59, 1, False)
    LAST_HP_DRIVER_ERRORS = (0xFF, 1, False)


class TicCommand(int, Enum):
    """
    Command codes for Polulu Tic Stepper Motor Controller.
    https://www.pololu.com/docs/0J71/8
    """

    SET_TARGET_POSITION = 0xE0
    SET_TARGET_VELOCITY = 0xE3
    HALT_AND_SET_POSITION = 0xEC
    HALT_AND_HOLD = 0x89
    GO_HOME = 0x97
    RESET_COMMAND_TIMEOUT = 0x8C
    DEENERGIZE = 0x86
    ENERGIZE = 0x85
    EXIT_SAFE_START = 0x83
    ENTER_SAFE_START = 0x8F
    RESET = 0xB
    CLEAR_DRIVER_ERROR = 0x8A
    SET_MAX_SPEED = 0xE6
    SET_STARTING_SPEED = 0xE5
    SET_MAX_ACCELERATION = 0xEA
    SET_MAX_DECELERATION = 0xE9
    SET_STEP_MODE = 0x94
    SET_CURRENT_LIMIT = 0x91
    SET_DECAY_MODE = 0x92
    SET_AGC_OPTION = 0x98
    GET_VARIABLE = 0xA1
    GET_VARIABLE_AND_CLEAR_ERRORS_OCCURRED = 0xA2
    GET_SETTING = 0xA8
    SET_SETTING = 0x13
    REINITIALIZE = 0x10
    START_BOOTLOADER = 0xFF


class TicDirection(int, Enum):
    """Possible directions for homing"""

    REVERSE = 0
    FORWARD = 1


class Tic(Stage):
    """
    Very basic driver class for Polulu Tic Stepper Motor controller, connected in USB.

    :ivar poll_interval: Interval between successive state polling for some
     long motor operations.
    """

    def __init__(self):
        super().__init__()
        dev = usb.core.find(idVendor=0x1FFB, idProduct=0x00B5)
        if isinstance(dev, usb.core.Device):
            self.dev = dev
        else:
            raise ConnectionFailure("Tic stepper motor not found.")
        self.dev.set_configuration()
        self.energize()
        self.poll_interval = 0.1

    def quick(self, command: TicCommand):
        """
        Send a quick command with no data.

        :param command: Command.
        """
        self.dev.ctrl_transfer(0x40, command, 0, 0, 0)

    def write_7(self, command: TicCommand, data: int):
        """
        Write 7 bits.

        :param command: Command.
        :param data: Value to be written.
        """
        self.dev.ctrl_transfer(0x40, command, data, 0, 0)

    def write_32(self, command: TicCommand, data: int):
        """
        Write 32 bits.

        :param command: Command code.
        :param data: Value to be written.
        """
        self.dev.ctrl_transfer(0x40, command, data & 0xFFFF, data >> 16, 0)

    def block_read(self, command: TicCommand, offset, length) -> bytes:
        """
        Read data from the device.

        :param command: Command code.
        :param offset: Data offset.
        :param length: Data length.
        """
        return bytes(self.dev.ctrl_transfer(0xC0, command, 0, offset, length))

    def set_setting(self, command: TicCommand, data, offset):
        """
        Set setting data.

        :param command: Command code.
        :param data: Value to be written.
        :param offset: Write offset.
        """
        self.dev.ctrl_transfer(0x40, command, data, offset, 0)

    def energize(self):
        self.quick(TicCommand.ENERGIZE)

    def deenergize(self):
        self.quick(TicCommand.DEENERGIZE)

    def reset(self):
        self.quick(TicCommand.RESET)

    def exit_safe_start(self):
        self.quick(TicCommand.EXIT_SAFE_START)

    def home(self, wait=False):
        """Triggers a Home command.

        :param wait: Optionally waits for move operation to be done."""
        self.go_home(TicDirection.REVERSE, False)
        if wait:
            self.wait_move_finished()

    def go_home(self, direction: TicDirection, wait: bool = True):
        """
        Run the homing procedure.
        :param direction: Homing direction.
        :param wait: If True, wait for homing procedure end.
        """
        self.exit_safe_start()
        self.write_7(TicCommand.GO_HOME, direction)
        if wait:
            while self.get_variable(TicVariable.MISC_FLAGS) & (1 << 4):
                self.exit_safe_start()
                sleep(self.poll_interval)

    def set_target_position(self, pos: int):
        self.write_32(TicCommand.SET_TARGET_POSITION, pos)

    def set_target_velocity(self, velocity: int):
        self.write_32(TicCommand.SET_TARGET_VELOCITY, velocity)

    def get_variable(self, variable: TicVariable) -> int:
        offset, length, signed = variable.value
        return int.from_bytes(
            self.block_read(TicCommand.GET_VARIABLE, offset, length),
            "little",
            signed=signed,
        )

    @property
    def position(self) -> Vector:
        """
        Motor position, in steps.

        :getter: Returns current target position.
        :setter: Set target position and wait until position is reached.
        """
        return Vector(self.get_variable(TicVariable.CURRENT_POSITION))

    @position.setter
    def position(self, value: Vector):
        # To check dimension and range of the given value
        pos_setter = Stage.position.fset
        assert pos_setter is not None
        pos_setter(self, value)

        self.target_position = value.x
        while self.position.x != value.x:
            sleep(self.poll_interval)
            self.exit_safe_start()

    @property
    def target_position(self) -> int:
        return self.get_variable(TicVariable.TARGET_POSITION)

    @target_position.setter
    def target_position(self, value: int):
        self.exit_safe_start()
        self.set_target_position(value)

    @property
    def is_moving(self) -> bool:
        """
        Because there is no specific command to get the current moving state of the Tic, and
        the positioning/homing is always blocking, this function returns False in all cases.
        """
        return False



from PyQt6.QtWidgets import QApplication, QStyleFactory
import sys
from PyQt6.QtGui import QIcon, QPalette, QColor
from .util import resource_path
from .gui import StageWindow

app = QApplication(sys.argv)

app.setApplicationName("PyStages")
app.setWindowIcon(QIcon(resource_path(":/icons/logo.png")))
app.setStyle(QStyleFactory.create("Fusion"))
palette = QPalette()
palette.setColor(QPalette.ColorRole.Window, QColor(25, 25, 25))
palette.setColor(QPalette.ColorRole.WindowText, QColor(240, 240, 240))
palette.setColor(QPalette.ColorRole.Base, QColor(40, 40, 40))
palette.setColor(QPalette.ColorRole.AlternateBase, QColor(255, 0, 0))
palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(25, 25, 25))
palette.setColor(QPalette.ColorRole.ToolTipText, QColor(255, 255, 255))
palette.setColor(QPalette.ColorRole.Text, QColor(200, 200, 200))
palette.setColor(QPalette.ColorRole.Button, QColor(40, 40, 40))
palette.setColor(
    QPalette.ColorGroup.Disabled, QPalette.ColorRole.Button, QColor(30, 30, 30)
)
palette.setColor(QPalette.ColorRole.ButtonText, QColor(200, 200, 200))
palette.setColor(
    QPalette.ColorGroup.Disabled, QPalette.ColorRole.ButtonText, QColor(100, 100, 100)
)
palette.setColor(QPalette.ColorRole.BrightText, QColor(255, 0, 0))
palette.setColor(QPalette.ColorRole.Link, QColor(255, 0, 0))
palette.setColor(QPalette.ColorRole.Highlight, QColor(40, 120, 233))
palette.setColor(QPalette.ColorRole.HighlightedText, QColor(255, 255, 255))
app.setPalette(palette)

win = StageWindow()
win.show()
sys.exit(app.exec())



import os

__dirname = os.path.dirname(os.path.realpath(__file__))


def resource_path(path: str) -> str:
    """
    Transforms a .":/path/to/file" path to the relative path from the main script
    """
    if not path.startswith(":/"):
        return path
    return os.path.join(__dirname, path[2:])
